@@ -2,12 +2,15 @@ import bisect
 import datetime
 import io
 import logging
-import os
 import xml.etree.ElementTree as ET  # noqa: N817
+from pathlib import Path
 
 import numpy as np
 
-ns = {"fews": "http://www.wldelft.nl/fews", "pi": "http://www.wldelft.nl/fews/PI"}
+from .common import NAMESPACES, TimeseriesId
+
+
+ns = NAMESPACES
 
 
 class Diag:
@@ -21,16 +24,13 @@ class Diag:
     INFO = 1 << 3
     DEBUG = 1 << 4
 
-    def __init__(self, folder, basename="diag"):
+    def __init__(self, path):
         """
         Parse diag file.
 
-        :param folder:   Folder in which diag.xml is found or to be created.
-        :param basename: Alternative basename for the diagnostics XML file.
+        :param path: Path to diagnostics XML file.
         """
-        self.__path_xml = os.path.join(folder, basename + ".xml")
-
-        self.__tree = ET.parse(self.__path_xml)
+        self.__tree = ET.parse(path)
         self.__xml_root = self.__tree.getroot()
 
     def get(self, level=ERROR_FATAL):
@@ -81,11 +81,10 @@ class DiagHandler(logging.Handler):
     PI diag file logging handler.
     """
 
-    def __init__(self, folder, basename="diag", level=logging.NOTSET):
-        super(DiagHandler, self).__init__(level=level)
-
-        self.__path_xml = os.path.join(folder, basename + ".xml")
+    def __init__(self, path, level=logging.NOTSET):
+        super().__init__(level=level)
 
+        self.__path_xml = path
         try:
             self.__tree = ET.parse(self.__path_xml)
             self.__xml_root = self.__tree.getroot()
@@ -121,19 +120,16 @@ class DiagHandler(logging.Handler):
 
 class ParameterConfig:
     """
-    rtcParameterConfig wrapper.
+    PI-XML parameter config wrapper.
     """
 
-    def __init__(self, folder, basename):
+    def __init__(self, path):
         """
-        Parses a rtcParameterConfig file.
+        Parses a PI-XML parameter config file.
 
-        :param folder:    Folder in which the parameter configuration file is located.
-        :param basename:  Basename of the parameter configuration file (e.g, 'rtcParameterConfig').
+        :param path: Path to parameter configuration file.
         """
-        if os.path.splitext(basename)[1] != ".xml":
-            basename = basename + ".xml"
-        self.__path_xml = os.path.join(folder, basename)
+        self.__path_xml = path
 
         self.__tree = ET.parse(self.__path_xml)
         self.__xml_root = self.__tree.getroot()
@@ -214,41 +210,17 @@ class ParameterConfig:
 
         raise KeyError("No such parameter ({}, {})".format(group_id, parameter_id))
 
-    def write(self, folder=None, basename=None):
+    def write(self, path=None):
         """
         Writes the parameter configuration to a file.
         Default behaviour is to overwrite original file.
 
-        :param path:     Optional alternative destination folder
-        :param basename: Optional alternative basename of the file (e.g, 'rtcParameterConfig')
+        :param path: Optional alternative location of export file
         """
 
         # No path changes- overwrite original file
-        if folder is None and basename is None:
-            path = self.path
-
-        # We need to reconstruct the path
-        else:
-            # Determine folder
-            if folder is not None:
-                if not os.path.exists(folder):
-                    # Make sure folder exists
-                    raise FileNotFoundError("Folder not found: {}".format(folder))
-            else:
-                # Reuse folder of original file
-                folder = os.path.dirname(self.path)
-
-            # Determine basename
-            if basename is not None:
-                # Make sure basename ends in '.xml'
-                if os.path.splitext(basename)[1] != ".xml":
-                    basename = basename + ".xml"
-            else:
-                # Reuse basename of original file
-                basename = os.path.split(self.path)[1]
-
-            # Construct path
-            path = os.path.join(folder, basename)
+        if path is None:
+            path = self.__path_xml
 
         self.__tree.write(path)
 
@@ -345,9 +317,7 @@ class Timeseries:
 
     def __init__(
         self,
-        data_config,
-        folder,
-        basename,
+        path,
         binary=True,
         pi_validate_times=False,
         make_new_file=False,
@@ -355,21 +325,14 @@ class Timeseries:
         """
         Load the timeseries from disk.
 
-        :param data_config:       A :class:`DataConfig` object.
-        :param folder:            The folder in which the time series is located.
-        :param basename:          The basename of the time series file.
+        :param path:              Path to time series file.
         :param binary:            True if the time series data is stored in a separate binary file.
                                   Default is ``True``.
         :param pi_validate_times: Check consistency of times.  Default is ``False``.
         :param make_new_file:     Make new XML object which can be filled and written to a new file.
                                   Default is ``False``.
         """
-        self.__data_config = data_config
-
-        self.__folder = folder
-        self.__basename = basename
-
-        self.__path_xml = os.path.join(self.__folder, basename + ".xml")
+        self.__path_xml = path
 
         self.__internal_dtype = np.float64
         self.__pi_dtype = np.float32
@@ -413,14 +376,14 @@ class Timeseries:
             for series in self.__xml_root.findall("pi:series", ns):
                 header = series.find("pi:header", ns)
 
-                variable = self.__data_config.variable(header)
+                timeseries_id = TimeseriesId.from_xml_element(header, "pi")
 
                 try:
                     dt = self.__parse_time_step(header.find("pi:timeStep", ns))
                 except ValueError:
                     raise Exception(
-                        "PI: Multiplier of time step of variable {} "
-                        "must be a positive integer per the PI schema.".format(variable)
+                        "PI: Multiplier of time step of {} "
+                        "must be a positive integer per the PI schema.".format(timeseries_id)
                     )
                 if self.__dt is None:
                     self.__dt = dt
@@ -436,9 +399,7 @@ class Timeseries:
                             self.__start_datetime = start_datetime
                 except (AttributeError, ValueError):
                     raise Exception(
-                        "PI: Variable {} in {} has no startDate.".format(
-                            variable, os.path.join(self.__folder, basename + ".xml")
-                        )
+                        "PI: {} in {} has no startDate.".format(timeseries_id, self.__path_xml)
                     )
 
                 try:
@@ -450,9 +411,7 @@ class Timeseries:
                             self.__end_datetime = end_datetime
                 except (AttributeError, ValueError):
                     raise Exception(
-                        "PI: Variable {} in {} has no endDate.".format(
-                            variable, os.path.join(self.__folder, basename + ".xml")
-                        )
+                        "PI: {} in {} has no endDate.".format(timeseries_id, self.__path_xml)
                     )
 
                 el = header.find("pi:forecastDate", ns)
@@ -530,7 +489,7 @@ class Timeseries:
             for series in self.__xml_root.findall("pi:series", ns):
                 header = series.find("pi:header", ns)
 
-                variable = self.__data_config.variable(header)
+                timeseries_id = TimeseriesId.from_xml_element(header, "pi")
 
                 dt = self.__parse_time_step(header.find("pi:timeStep", ns))
                 start_datetime = self.__parse_date_time(header.find("pi:startDate", ns))
@@ -571,34 +530,36 @@ class Timeseries:
 
                 if self.__binary:
                     if f is not None:
-                        self.__values[ensemble_member][variable] = np.fromfile(
+                        self.__values[ensemble_member][timeseries_id] = np.fromfile(
                             f, count=n_values, dtype=self.__pi_dtype
                         )
                     else:
-                        self.__values[ensemble_member][variable] = np.empty(
+                        self.__values[ensemble_member][timeseries_id] = np.empty(
                             n_values, dtype=self.__internal_dtype
                         )
-                        self.__values[ensemble_member][variable].fill(np.nan)
+                        self.__values[ensemble_member][timeseries_id].fill(np.nan)
 
                 else:
                     events = series.findall("pi:event", ns)
 
-                    self.__values[ensemble_member][variable] = np.empty(
+                    self.__values[ensemble_member][timeseries_id] = np.empty(
                         n_values, dtype=self.__internal_dtype
                     )
-                    self.__values[ensemble_member][variable].fill(np.nan)
+                    self.__values[ensemble_member][timeseries_id].fill(np.nan)
                     # This assumes that start_datetime equals the datetime of the
                     # first value (which should be the case).
                     for i in range(min(n_values, len(events))):
-                        self.__values[ensemble_member][variable][i] = float(events[i].get("value"))
+                        self.__values[ensemble_member][timeseries_id][i] = float(
+                            events[i].get("value")
+                        )
 
                 miss_val = float(header.find("pi:missVal", ns).text)
-                self.__values[ensemble_member][variable][
-                    self.__values[ensemble_member][variable] == miss_val
+                self.__values[ensemble_member][timeseries_id][
+                    self.__values[ensemble_member][timeseries_id] == miss_val
                 ] = np.nan
 
                 unit = header.find("pi:units", ns).text
-                self.set_unit(variable, unit=unit, ensemble_member=ensemble_member)
+                self.set_unit(timeseries_id, unit=unit, ensemble_member=ensemble_member)
 
                 # Prepend empty space, if start_datetime > self.__start_datetime.
                 if start_datetime > self.__start_datetime:
@@ -624,8 +585,8 @@ class Timeseries:
                         )
 
                     filler.fill(np.nan)
-                    self.__values[ensemble_member][variable] = np.hstack(
-                        (filler, self.__values[ensemble_member][variable])
+                    self.__values[ensemble_member][timeseries_id] = np.hstack(
+                        (filler, self.__values[ensemble_member][timeseries_id])
                     )
 
                 # Append empty space, if end_datetime < self.__end_datetime
@@ -652,16 +613,16 @@ class Timeseries:
                         )
 
                     filler.fill(np.nan)
-                    self.__values[ensemble_member][variable] = np.hstack(
-                        (self.__values[ensemble_member][variable], filler)
+                    self.__values[ensemble_member][timeseries_id] = np.hstack(
+                        (self.__values[ensemble_member][timeseries_id], filler)
                     )
 
                 if make_virtual_ensemble:
                     # Make references to the original input series for the virtual
                     # ensemble members.
                     for i in range(1, self.ensemble_size):
-                        self.__values[i][variable] = self.__values[0][variable]
-                        self.set_unit(variable, unit=unit, ensemble_member=i)
+                        self.__values[i][timeseries_id] = self.__values[0][timeseries_id]
+                        self.set_unit(timeseries_id, unit=unit, ensemble_member=i)
 
             if not self.__dt:
                 # Remove time values outside the start/end datetimes.
@@ -717,9 +678,7 @@ class Timeseries:
         rounding = (seconds + round_to / 2) // round_to * round_to
         return dt + datetime.timedelta(0, rounding - seconds, -dt.microsecond)
 
-    def __add_header(
-        self, variable, location_parameter_id, ensemble_member=0, miss_val=-999, unit="unit_unknown"
-    ):
+    def __add_header(self, timeseries_id, ensemble_member=0, miss_val=-999, unit="unit_unknown"):
         """
         Add a timeseries header to the timeseries object.
         """
@@ -742,13 +701,13 @@ class Timeseries:
         ]
         header_element_texts = [
             "instantaneous",
-            location_parameter_id.location_id,
-            location_parameter_id.parameter_id,
+            timeseries_id.location_id,
+            timeseries_id.parameter_id,
             "",
             "",
             "",
             str(miss_val),
-            location_parameter_id.location_id,
+            timeseries_id.location_id,
             unit,
             now.strftime("%Y-%m-%d"),
             now.strftime("%H:%M:%S"),
@@ -761,10 +720,10 @@ class Timeseries:
         if self.contains_ensemble:
             header_elements.insert(3, "ensembleMemberIndex")
             header_element_texts.insert(3, str(ensemble_member))
-        if len(location_parameter_id.qualifier_id) > 0:
+        if len(timeseries_id.qualifier_id) > 0:
             # Track relative index to preserve original ordering of qualifier ID's
             i = 0
-            for qualifier_id in location_parameter_id.qualifier_id:
+            for qualifier_id in timeseries_id.qualifier_id:
                 header_elements.insert(3, "qualifierId")
                 header_element_texts.insert(3 + i, qualifier_id)
                 i += 1
@@ -822,12 +781,10 @@ class Timeseries:
             self.__reset_xml_tree()
 
             for ensemble_member in range(len(self.__values)):
-                for variable in sorted(self.__values[ensemble_member].keys()):
-                    location_parameter_id = self.__data_config.pi_variable_ids(variable)
-                    unit = self.get_unit(variable, ensemble_member)
+                for timeseries_id in sorted(self.__values[ensemble_member].keys()):
+                    unit = self.get_unit(timeseries_id, ensemble_member)
                     self.__add_header(
-                        variable,
-                        location_parameter_id,
+                        timeseries_id,
                         ensemble_member=ensemble_member,
                         miss_val=-999,
                         unit=unit,
@@ -836,6 +793,7 @@ class Timeseries:
         for ensemble_member in range(len(self.__values)):
             for series in self.__xml_root.findall("pi:series", ns):
                 header = series.find("pi:header", ns)
+                timeseries_id = TimeseriesId.from_xml_element(header, "pi")
 
                 # First check ensembleMemberIndex, to see if it is the correct one.
                 el = header.find("pi:ensembleMemberIndex", ns)
@@ -853,14 +811,12 @@ class Timeseries:
                 el.set("date", self.__end_datetime.strftime("%Y-%m-%d"))
                 el.set("time", self.__end_datetime.strftime("%H:%M:%S"))
 
-                variable = self.__data_config.variable(header)
-
                 miss_val = float(header.find("pi:missVal", ns).text)
-                values = self.__values[ensemble_member][variable]
+                values = self.__values[ensemble_member][timeseries_id]
 
                 # Update the header, which may have changed
                 el = header.find("pi:units", ns)
-                el.text = self.get_unit(variable, ensemble_member)
+                el.text = self.get_unit(timeseries_id, ensemble_member)
 
                 # No values to be written, so the entire element is removed from
                 # the XML, and the loop restarts.
@@ -1013,55 +969,55 @@ class Timeseries:
     def timezone(self, timezone):
         self.__timezone = timezone
 
-    def get(self, variable, ensemble_member=0):
+    def get(self, timeseries_id, ensemble_member=0):
         """
         Look up a time series.
 
-        :param variable:        Time series ID.
+        :param timeseries_id:        Time series ID.
         :param ensemble_member: Ensemble member index.
 
         :returns: A :class:`Timeseries` object.
         """
-        return self.__values[ensemble_member][variable]
+        return self.__values[ensemble_member][timeseries_id]
 
-    def set(self, variable, new_values, unit=None, ensemble_member=0):
+    def set(self, timeseries_id, new_values, unit=None, ensemble_member=0):
         """
         Fill a time series with new values, and set the unit.
 
-        :param variable:        Time series ID.
+        :param timeseries_id:        Time series ID.
         :param new_values:      List of new values.
         :param unit:            Unit.
         :param ensemble_member: Ensemble member index.
         """
-        self.__values[ensemble_member][variable] = new_values
+        self.__values[ensemble_member][timeseries_id] = new_values
         if unit is None:
-            unit = self.get_unit(variable, ensemble_member)
-        self.set_unit(variable, unit, ensemble_member)
+            unit = self.get_unit(timeseries_id, ensemble_member)
+        self.set_unit(timeseries_id, unit, ensemble_member)
 
-    def get_unit(self, variable, ensemble_member=0):
+    def get_unit(self, timeseries_id, ensemble_member=0):
         """
         Look up the unit of a time series.
 
-        :param variable:        Time series ID.
+        :param timeseries_id:        Time series ID.
         :param ensemble_member: Ensemble member index.
 
         :returns: A :string: containing the unit.
-                  If not set for the variable, returns 'unit_unknown'.
+                  If not set for the timeseries_id, returns 'unit_unknown'.
         """
         try:
-            return self.__units[ensemble_member][variable]
+            return self.__units[ensemble_member][timeseries_id]
         except KeyError:
             return "unit_unknown"
 
-    def set_unit(self, variable, unit, ensemble_member=0):
+    def set_unit(self, timeseries_id, unit, ensemble_member=0):
         """
         Set the unit of a time series.
 
-        :param variable:        Time series ID.
+        :param timeseries_id:        Time series ID.
         :param unit:            Unit.
         :param ensemble_member: Ensemble member index.
         """
-        self.__units[ensemble_member][variable] = unit
+        self.__units[ensemble_member][timeseries_id] = unit
 
     def resize(self, start_datetime, end_datetime):
         """
@@ -1149,7 +1105,7 @@ class Timeseries:
         """
         The path for the binary data .bin file.
         """
-        return os.path.join(self.__folder, self.__basename + ".bin")
+        return Path(self.__path_xml).with_suffix(".bin")
 
     def items(self, ensemble_member=0):
         """
@@ -1158,3 +1114,27 @@ class Timeseries:
         """
         for key in self.__values[ensemble_member].keys():
             yield key, self.get(key, ensemble_member)
+
+    def to_dataframe(self, ensemble_member=0):
+        import pandas as pd
+
+        if ensemble_member is None:
+            dfs = []
+            for ensemble_member in range(self.ensemble_size):
+                df = self.to_dataframe(ensemble_member)
+                df = (
+                    df.assign(ensemble_member=ensemble_member)
+                    .set_index("ensemble_member", append=True)
+                    .unstack("ensemble_member")
+                )
+                dfs.append(df)
+            df = pd.concat(dfs, axis=1)
+        else:
+            df = pd.DataFrame.from_dict(dict(self.items(ensemble_member)))
+            df.index = self.times
+            col_tuples = [(x.location_id, x.parameter_id, x.qualifier_ids) for x in df.columns]
+            df.columns = pd.MultiIndex.from_tuples(
+                col_tuples, names=["location_id", "parameter_id", "qualifier_ids"]
+            )
+
+        return df
